@@ -3,6 +3,9 @@
  * License.....: MIT
  */
 
+#ifndef _FOLDER_H
+#define _FOLDER_H
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <errno.h>
@@ -24,8 +27,6 @@
 #define DOT_HASHCAT     ".hashcat"
 #define SESSIONS_FOLDER "sessions"
 
-int sort_by_stringptr (const void *p1, const void *p2);
-
 int count_dictionaries (char **dictionary_files);
 
 char *first_file_in_directory (const char *path);
@@ -36,3 +37,5 @@ int  folder_config_init    (hashcat_ctx_t *hashcat_ctx, MAYBE_UNUSED const char
 void folder_config_destroy (hashcat_ctx_t *hashcat_ctx);
 
 int hc_mkdir (const char *name, MAYBE_UNUSED const int mode);
+
+#endif // _FOLDER_H
