@@ -209,6 +209,7 @@ typedef enum status_rc
   STATUS_ABORTED_CHECKPOINT = 10,
   STATUS_ABORTED_RUNTIME    = 11,
   STATUS_ERROR              = 13,
+  STATUS_ABORTED_FINISH     = 14,
 
 } status_rc_t;
 
@@ -239,6 +240,7 @@ typedef enum attack_mode
   ATTACK_MODE_TABLE     = 5,
   ATTACK_MODE_HYBRID1   = 6,
   ATTACK_MODE_HYBRID2   = 7,
+  ATTACK_MODE_ASSOCIATION   = 9,
   ATTACK_MODE_NONE      = 100
 
 } attack_mode_t;
@@ -256,12 +258,14 @@ typedef enum kern_run
 {
   KERN_RUN_1      = 1000,
   KERN_RUN_12     = 1500,
+  KERN_RUN_2P     = 1999,
   KERN_RUN_2      = 2000,
   KERN_RUN_2E     = 2001,
   KERN_RUN_23     = 2500,
   KERN_RUN_3      = 3000,
   KERN_RUN_4      = 4000,
   KERN_RUN_INIT2  = 5000,
+  KERN_RUN_LOOP2P = 5999,
   KERN_RUN_LOOP2  = 6000,
   KERN_RUN_AUX1   = 7001,
   KERN_RUN_AUX2   = 7002,
@@ -354,27 +358,29 @@ typedef enum salt_type
 
 typedef enum opti_type
 {
-  OPTI_TYPE_OPTIMIZED_KERNEL    = (1 <<  0),
-  OPTI_TYPE_ZERO_BYTE           = (1 <<  1),
-  OPTI_TYPE_PRECOMPUTE_INIT     = (1 <<  2),
-  OPTI_TYPE_MEET_IN_MIDDLE      = (1 <<  3),
-  OPTI_TYPE_EARLY_SKIP          = (1 <<  4),
-  OPTI_TYPE_NOT_SALTED          = (1 <<  5),
-  OPTI_TYPE_NOT_ITERATED        = (1 <<  6),
-  OPTI_TYPE_PREPENDED_SALT      = (1 <<  7),
-  OPTI_TYPE_APPENDED_SALT       = (1 <<  8),
-  OPTI_TYPE_SINGLE_HASH         = (1 <<  9),
-  OPTI_TYPE_SINGLE_SALT         = (1 << 10),
-  OPTI_TYPE_BRUTE_FORCE         = (1 << 11),
-  OPTI_TYPE_RAW_HASH            = (1 << 12),
-  OPTI_TYPE_SLOW_HASH_SIMD_INIT = (1 << 13),
-  OPTI_TYPE_SLOW_HASH_SIMD_LOOP = (1 << 14),
-  OPTI_TYPE_SLOW_HASH_SIMD_COMP = (1 << 15),
-  OPTI_TYPE_USES_BITS_8         = (1 << 16),
-  OPTI_TYPE_USES_BITS_16        = (1 << 17),
-  OPTI_TYPE_USES_BITS_32        = (1 << 18),
-  OPTI_TYPE_USES_BITS_64        = (1 << 19),
-  OPTI_TYPE_REGISTER_LIMIT      = (1 << 20), // We'll limit the register count to 128
+  OPTI_TYPE_OPTIMIZED_KERNEL      = (1 <<  0),
+  OPTI_TYPE_ZERO_BYTE             = (1 <<  1),
+  OPTI_TYPE_PRECOMPUTE_INIT       = (1 <<  2),
+  OPTI_TYPE_MEET_IN_MIDDLE        = (1 <<  3),
+  OPTI_TYPE_EARLY_SKIP            = (1 <<  4),
+  OPTI_TYPE_NOT_SALTED            = (1 <<  5),
+  OPTI_TYPE_NOT_ITERATED          = (1 <<  6),
+  OPTI_TYPE_PREPENDED_SALT        = (1 <<  7),
+  OPTI_TYPE_APPENDED_SALT         = (1 <<  8),
+  OPTI_TYPE_SINGLE_HASH           = (1 <<  9),
+  OPTI_TYPE_SINGLE_SALT           = (1 << 10),
+  OPTI_TYPE_BRUTE_FORCE           = (1 << 11),
+  OPTI_TYPE_RAW_HASH              = (1 << 12),
+  OPTI_TYPE_SLOW_HASH_SIMD_INIT   = (1 << 13),
+  OPTI_TYPE_SLOW_HASH_SIMD_LOOP   = (1 << 14),
+  OPTI_TYPE_SLOW_HASH_SIMD_COMP   = (1 << 15),
+  OPTI_TYPE_USES_BITS_8           = (1 << 16),
+  OPTI_TYPE_USES_BITS_16          = (1 << 17),
+  OPTI_TYPE_USES_BITS_32          = (1 << 18),
+  OPTI_TYPE_USES_BITS_64          = (1 << 19),
+  OPTI_TYPE_REGISTER_LIMIT        = (1 << 20), // We'll limit the register count to 128
+  OPTI_TYPE_SLOW_HASH_SIMD_INIT2  = (1 << 21),
+  OPTI_TYPE_SLOW_HASH_SIMD_LOOP2  = (1 << 22),
 
 } opti_type_t;
 
@@ -409,26 +415,33 @@ typedef enum opts_type
   OPTS_TYPE_ST_BASE64         = (1ULL << 26),
   OPTS_TYPE_HASH_COPY         = (1ULL << 28),
   OPTS_TYPE_HASH_SPLIT        = (1ULL << 29),
-  OPTS_TYPE_LOOP_EXTENDED     = (1ULL << 30), // a kernel which is called each time normal _loop kernel finished.
+  OPTS_TYPE_LOOP_PREPARE      = (1ULL << 30), // a kernel which is called each time before _loop kernel started.
+                                              // like a hook12 kernel but without extra buffers.
+  OPTS_TYPE_LOOP_EXTENDED     = (1ULL << 31), // a kernel which is called each time normal _loop kernel finished.
                                               // but unlike a hook kernel this kernel is called for every _loop iteration offset
-  OPTS_TYPE_HOOK12            = (1ULL << 31),
-  OPTS_TYPE_HOOK23            = (1ULL << 32),
-  OPTS_TYPE_INIT2             = (1ULL << 33),
-  OPTS_TYPE_LOOP2             = (1ULL << 34),
-  OPTS_TYPE_AUX1              = (1ULL << 35),
-  OPTS_TYPE_AUX2              = (1ULL << 36),
-  OPTS_TYPE_AUX3              = (1ULL << 37),
-  OPTS_TYPE_AUX4              = (1ULL << 38),
-  OPTS_TYPE_BINARY_HASHFILE   = (1ULL << 39),
-  OPTS_TYPE_PT_ADD06          = (1ULL << 41),
-  OPTS_TYPE_KEYBOARD_MAPPING  = (1ULL << 42),
-  OPTS_TYPE_DEEP_COMP_KERNEL  = (1ULL << 43), // if we have to iterate through each hash inside the comp kernel, for example if each hash has to be decrypted separately
-  OPTS_TYPE_TM_KERNEL         = (1ULL << 44),
-  OPTS_TYPE_SUGGEST_KG        = (1ULL << 45), // suggest keep guessing for modules the user maybe wants to use --keep-guessing
-  OPTS_TYPE_COPY_TMPS         = (1ULL << 46), // if we want to use data from tmps buffer (for example get the PMK in WPA)
-  OPTS_TYPE_POTFILE_NOPASS    = (1ULL << 47), // sometimes the password should not be printed to potfile
-  OPTS_TYPE_DYNAMIC_SHARED    = (1ULL << 48), // use dynamic shared memory (note: needs special kernel changes)
-  OPTS_TYPE_SELF_TEST_DISABLE = (1ULL << 49), // some algos use JiT in combinations with a salt or create too much startup time
+  OPTS_TYPE_HOOK12            = (1ULL << 32),
+  OPTS_TYPE_HOOK23            = (1ULL << 33),
+  OPTS_TYPE_INIT2             = (1ULL << 34),
+  OPTS_TYPE_LOOP2_PREPARE     = (1ULL << 35), // same as OPTS_TYPE_LOOP_PREPARE but for loop2 kernel
+  OPTS_TYPE_LOOP2             = (1ULL << 36),
+  OPTS_TYPE_AUX1              = (1ULL << 37),
+  OPTS_TYPE_AUX2              = (1ULL << 38),
+  OPTS_TYPE_AUX3              = (1ULL << 39),
+  OPTS_TYPE_AUX4              = (1ULL << 40),
+  OPTS_TYPE_BINARY_HASHFILE   = (1ULL << 41),
+  OPTS_TYPE_BINARY_HASHFILE_OPTIONAL
+                              = (1ULL << 42), // this allows us to not enforce the use of a binary file. requires OPTS_TYPE_BINARY_HASHFILE set to be effective.
+  OPTS_TYPE_PT_ADD06          = (1ULL << 43),
+  OPTS_TYPE_KEYBOARD_MAPPING  = (1ULL << 44),
+  OPTS_TYPE_DEEP_COMP_KERNEL  = (1ULL << 45), // if we have to iterate through each hash inside the comp kernel, for example if each hash has to be decrypted separately
+  OPTS_TYPE_TM_KERNEL         = (1ULL << 46),
+  OPTS_TYPE_SUGGEST_KG        = (1ULL << 47), // suggest keep guessing for modules the user maybe wants to use --keep-guessing
+  OPTS_TYPE_COPY_TMPS         = (1ULL << 48), // if we want to use data from tmps buffer (for example get the PMK in WPA)
+  OPTS_TYPE_POTFILE_NOPASS    = (1ULL << 49), // sometimes the password should not be printed to potfile
+  OPTS_TYPE_DYNAMIC_SHARED    = (1ULL << 50), // use dynamic shared memory (note: needs special kernel changes)
+  OPTS_TYPE_SELF_TEST_DISABLE = (1ULL << 51), // some algos use JiT in combinations with a salt or create too much startup time
+  OPTS_TYPE_MP_MULTI_DISABLE  = (1ULL << 52), // do not multiply the kernel-accel with the multiprocessor count per device to allow more fine-tuned workload settings
+  OPTS_TYPE_NATIVE_THREADS    = (1ULL << 53), // forces "native" thread count: CPU=1, GPU-Intel=8, GPU-AMD=64 (wavefront), GPU-NV=32 (warps)
 
 } opts_type_t;
 
@@ -539,6 +552,8 @@ typedef enum parser_rc
   PARSER_BLOCK_SIZE           = -39,
   PARSER_CIPHER               = -40,
   PARSER_FILE_SIZE            = -41,
+  PARSER_IV_LENGTH            = -42,
+  PARSER_CT_LENGTH            = -43,
   PARSER_HAVE_ERRNO           = -100,
   PARSER_UNKNOWN_ERROR        = -255
 
@@ -588,10 +603,10 @@ typedef enum user_options_defaults
   BRAIN_SESSION            = 0,
   #endif
   DEBUG_MODE               = 0,
-  EXAMPLE_HASHES           = false,
   FORCE                    = false,
   HWMON_DISABLE            = false,
   HWMON_TEMP_ABORT         = 90,
+  HASH_INFO                = false,
   HASH_MODE                = 0,
   HCCAPX_MESSAGE_PAIR      = 0,
   HEX_CHARSET              = false,
@@ -693,7 +708,7 @@ typedef enum user_options_map
   IDX_DEBUG_MODE                = 0xff11,
   IDX_ENCODING_FROM             = 0xff12,
   IDX_ENCODING_TO               = 0xff13,
-  IDX_EXAMPLE_HASHES            = 0xff14,
+  IDX_HASH_INFO                 = 0xff14,
   IDX_FORCE                     = 0xff15,
   IDX_HWMON_DISABLE             = 0xff16,
   IDX_HWMON_TEMP_ABORT          = 0xff17,
@@ -921,7 +936,7 @@ typedef struct hashes
 
 } hashes_t;
 
-struct hashconfig
+typedef struct hashconfig
 {
   char  separator;
 
@@ -945,6 +960,7 @@ struct hashconfig
   // sizes have to be size_t
 
   u64   esalt_size;
+  u64   hook_extra_param_size;
   u64   hook_salt_size;
   u64   tmp_size;
   u64   extra_tmp_size;
@@ -994,9 +1010,7 @@ struct hashconfig
   bool forced_jit_compile;
 
   u32 pwdump_column;
-};
-
-typedef struct hashconfig hashconfig_t;
+} hashconfig_t;
 
 typedef struct pw_pre
 {
@@ -1084,12 +1098,14 @@ typedef struct hc_device_param
 
   u32     kernel_wgs1;
   u32     kernel_wgs12;
+  u32     kernel_wgs2p;
   u32     kernel_wgs2;
   u32     kernel_wgs2e;
   u32     kernel_wgs23;
   u32     kernel_wgs3;
   u32     kernel_wgs4;
   u32     kernel_wgs_init2;
+  u32     kernel_wgs_loop2p;
   u32     kernel_wgs_loop2;
   u32     kernel_wgs_mp;
   u32     kernel_wgs_mp_l;
@@ -1106,12 +1122,14 @@ typedef struct hc_device_param
 
   u32     kernel_preferred_wgs_multiple1;
   u32     kernel_preferred_wgs_multiple12;
+  u32     kernel_preferred_wgs_multiple2p;
   u32     kernel_preferred_wgs_multiple2;
   u32     kernel_preferred_wgs_multiple2e;
   u32     kernel_preferred_wgs_multiple23;
   u32     kernel_preferred_wgs_multiple3;
   u32     kernel_preferred_wgs_multiple4;
   u32     kernel_preferred_wgs_multiple_init2;
+  u32     kernel_preferred_wgs_multiple_loop2p;
   u32     kernel_preferred_wgs_multiple_loop2;
   u32     kernel_preferred_wgs_multiple_mp;
   u32     kernel_preferred_wgs_multiple_mp_l;
@@ -1128,12 +1146,14 @@ typedef struct hc_device_param
 
   u64     kernel_local_mem_size1;
   u64     kernel_local_mem_size12;
+  u64     kernel_local_mem_size2p;
   u64     kernel_local_mem_size2;
   u64     kernel_local_mem_size2e;
   u64     kernel_local_mem_size23;
   u64     kernel_local_mem_size3;
   u64     kernel_local_mem_size4;
   u64     kernel_local_mem_size_init2;
+  u64     kernel_local_mem_size_loop2p;
   u64     kernel_local_mem_size_loop2;
   u64     kernel_local_mem_size_mp;
   u64     kernel_local_mem_size_mp_l;
@@ -1150,12 +1170,14 @@ typedef struct hc_device_param
 
   u64     kernel_dynamic_local_mem_size1;
   u64     kernel_dynamic_local_mem_size12;
+  u64     kernel_dynamic_local_mem_size2p;
   u64     kernel_dynamic_local_mem_size2;
   u64     kernel_dynamic_local_mem_size2e;
   u64     kernel_dynamic_local_mem_size23;
   u64     kernel_dynamic_local_mem_size3;
   u64     kernel_dynamic_local_mem_size4;
   u64     kernel_dynamic_local_mem_size_init2;
+  u64     kernel_dynamic_local_mem_size_loop2p;
   u64     kernel_dynamic_local_mem_size_loop2;
   u64     kernel_dynamic_local_mem_size_mp;
   u64     kernel_dynamic_local_mem_size_mp_l;
@@ -1263,11 +1285,13 @@ typedef struct hc_device_param
   // workaround cpu spinning
 
   double  exec_us_prev1[EXPECTED_ITERATIONS];
+  double  exec_us_prev2p[EXPECTED_ITERATIONS];
   double  exec_us_prev2[EXPECTED_ITERATIONS];
   double  exec_us_prev2e[EXPECTED_ITERATIONS];
   double  exec_us_prev3[EXPECTED_ITERATIONS];
   double  exec_us_prev4[EXPECTED_ITERATIONS];
   double  exec_us_prev_init2[EXPECTED_ITERATIONS];
+  double  exec_us_prev_loop2p[EXPECTED_ITERATIONS];
   double  exec_us_prev_loop2[EXPECTED_ITERATIONS];
   double  exec_us_prev_aux1[EXPECTED_ITERATIONS];
   double  exec_us_prev_aux2[EXPECTED_ITERATIONS];
@@ -1368,12 +1392,14 @@ typedef struct hc_device_param
 
   CUfunction        cuda_function1;
   CUfunction        cuda_function12;
+  CUfunction        cuda_function2p;
   CUfunction        cuda_function2;
   CUfunction        cuda_function2e;
   CUfunction        cuda_function23;
   CUfunction        cuda_function3;
   CUfunction        cuda_function4;
   CUfunction        cuda_function_init2;
+  CUfunction        cuda_function_loop2p;
   CUfunction        cuda_function_loop2;
   CUfunction        cuda_function_mp;
   CUfunction        cuda_function_mp_l;
@@ -1392,8 +1418,6 @@ typedef struct hc_device_param
   CUdeviceptr       cuda_d_pws_amp_buf;
   CUdeviceptr       cuda_d_pws_comp_buf;
   CUdeviceptr       cuda_d_pws_idx;
-  CUdeviceptr       cuda_d_words_buf_l;
-  CUdeviceptr       cuda_d_words_buf_r;
   CUdeviceptr       cuda_d_rules;
   CUdeviceptr       cuda_d_rules_c;
   CUdeviceptr       cuda_d_combs;
@@ -1452,12 +1476,14 @@ typedef struct hc_device_param
 
   cl_kernel         opencl_kernel1;
   cl_kernel         opencl_kernel12;
+  cl_kernel         opencl_kernel2p;
   cl_kernel         opencl_kernel2;
   cl_kernel         opencl_kernel2e;
   cl_kernel         opencl_kernel23;
   cl_kernel         opencl_kernel3;
   cl_kernel         opencl_kernel4;
   cl_kernel         opencl_kernel_init2;
+  cl_kernel         opencl_kernel_loop2p;
   cl_kernel         opencl_kernel_loop2;
   cl_kernel         opencl_kernel_mp;
   cl_kernel         opencl_kernel_mp_l;
@@ -1476,8 +1502,6 @@ typedef struct hc_device_param
   cl_mem            opencl_d_pws_amp_buf;
   cl_mem            opencl_d_pws_comp_buf;
   cl_mem            opencl_d_pws_idx;
-  cl_mem            opencl_d_words_buf_l;
-  cl_mem            opencl_d_words_buf_r;
   cl_mem            opencl_d_rules;
   cl_mem            opencl_d_rules_c;
   cl_mem            opencl_d_combs;
@@ -1554,6 +1578,9 @@ typedef struct backend_ctx
 
   // cuda
 
+  int                 rc_cuda_init;
+  int                 rc_nvrtc_init;
+
   int                 nvrtc_driver_version;
   int                 cuda_driver_version;
 
@@ -1923,6 +1950,7 @@ typedef struct user_options
   bool         workload_profile_chgd;
   bool         skip_chgd;
   bool         limit_chgd;
+  bool         scrypt_tmto_chgd;
 
   bool         advice_disable;
   bool         benchmark;
@@ -1931,9 +1959,9 @@ typedef struct user_options
   bool         brain_client;
   bool         brain_server;
   #endif
-  bool         example_hashes;
   bool         force;
   bool         hwmon_disable;
+  bool         hash_info;
   bool         hex_charset;
   bool         hex_salt;
   bool         hex_wordlist;
@@ -2321,6 +2349,7 @@ typedef struct status_ctx
   bool shutdown_outer;
 
   bool checkpoint_shutdown;
+  bool finish_shutdown;
 
   hc_thread_mutex_t mux_dispatcher;
   hc_thread_mutex_t mux_counter;
@@ -2438,6 +2467,8 @@ typedef struct module_ctx
 
   MODULE_INIT module_init;
 
+  void      **hook_extra_params; // free for module to use (for instance: library handles)
+
   u32         (*module_attack_exec)             (const hashconfig_t *, const user_options_t *, const user_options_extra_t *);
   void       *(*module_benchmark_esalt)         (const hashconfig_t *, const user_options_t *, const user_options_extra_t *);
   void       *(*module_benchmark_hook_salt)     (const hashconfig_t *, const user_options_t *, const user_options_extra_t *);
@@ -2503,8 +2534,12 @@ typedef struct module_ctx
   u32         (*module_deep_comp_kernel)        (const hashes_t *, const u32, const u32);
   int         (*module_hash_init_selftest)      (const hashconfig_t *, hash_t *);
 
-  void        (*module_hook12)                  (hc_device_param_t *, const void *, const u32, const u64);
-  void        (*module_hook23)                  (hc_device_param_t *, const void *, const u32, const u64);
+  u64         (*module_hook_extra_param_size)   (const hashconfig_t *, const user_options_t *, const user_options_extra_t *);
+  bool        (*module_hook_extra_param_init)   (const hashconfig_t *, const user_options_t *, const user_options_extra_t *, const folder_config_t *, const backend_ctx_t *, void *);
+  bool        (*module_hook_extra_param_term)   (const hashconfig_t *, const user_options_t *, const user_options_extra_t *, const folder_config_t *, const backend_ctx_t *, void *);
+
+  void        (*module_hook12)                  (hc_device_param_t *, const void *, const void *, const u32, const u64);
+  void        (*module_hook23)                  (hc_device_param_t *, const void *, const void *, const u32, const u64);
 
   int         (*module_build_plain_postprocess) (const hashconfig_t *, const hashes_t *, const void *, const u32 *, const size_t, const int, u32 *, const size_t);
 
@@ -2567,6 +2602,7 @@ typedef struct hook_thread_param
 
   hc_device_param_t *device_param;
 
+  void *hook_extra_param;
   void *hook_salts_buf;
 
   u32 salt_pos;
@@ -2599,8 +2635,6 @@ typedef struct token
 
 } token_t;
 
-#endif // _TYPES_H
-
 /**
  * hash category is relevant in usage.c (--help screen)
  */
@@ -2627,9 +2661,12 @@ typedef enum hash_category
   HASH_CATEGORY_OTP                     = 17,
   HASH_CATEGORY_PLAIN                   = 18,
   HASH_CATEGORY_FRAMEWORK               = 19,
+  HASH_CATEGORY_PRIVATE_KEY             = 20,
 
 } hash_category_t;
 
 // hash specific
 
 typedef aes_ctx AES_KEY;
+
+#endif // _TYPES_H
